!---------------------------------------------------------------------------------------
! PREP_SYNDATA: mod_save
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_save
!!
!! Purpose:    Saves the perturbed fluxes and observations to files.
!!
!---------------------------------------------------------------------------------------

module mod_save

  use mod_var
  use mod_settings
  use netcdf

  implicit none
  private

  public :: save_flux, save_obs, check

  contains

  ! --------------------------------------------------
  ! save_flux
  ! --------------------------------------------------
  !> save_flux
  !!
  !! Purpose:  Adds the flux perturbations for the
  !!           inversion domain to the prior fluxes.
  !!           Writes the global fluxes (with 
  !!           perturbation over inversion domain) to
  !!           a ncdf file.
  !! 
  ! --------------------------------------------------

  subroutine save_flux(config, files, flx, state)

    use mod_var
    use mod_settings
    use mod_dates
    use mod_strings

    implicit none

    type (config_t),                       intent (in)     :: config
    type (files_t),                        intent (in)     :: files
    real, dimension(nxgrid,nygrid,nflx),   intent (in out) :: flx
    real, dimension(nvar),                 intent (in out) :: state

    character(len=max_name_len)                            :: filename, lonname, latname, timename, varname
    character(len=4)                                       :: adate
    integer                                                :: lon_dimid, lat_dimid, time_dimid
    integer                                                :: lon_varid, lat_varid, time_varid, varid, ncid
    integer, dimension(3)                                  :: dimids
    character(len=*), parameter                            :: units = 'units'
    character(len=*), parameter                            :: dimunit = 'degrees'
    character(len=*), parameter                            :: timeunit = 'julian_day'
    character(len=*), parameter                            :: flxunit = 'kg/m2/h'
    integer                                                :: d, n, nd, nn, nt, ii, i1, i2 
    integer                                                :: ix, jy, ix1, ix2, jy1, jy2
    real                                                   :: frac, hh, hutc
    real, dimension(nxregrid,nyregrid,nflx)                :: flx_out
    real, dimension(nflx)                                  :: time
    real, dimension(:,:,:), allocatable                    :: tmp
    real(kind=8)                                           :: jd
    integer                                                :: yyyymmdd, hhmiss

    flx_out(:,:,:) = 0.

    ! coordinates to global domain
    ix1 = minloc(gbl_lon,dim=1,mask=abs(gbl_lon-rllx).eq.minval(abs(gbl_lon-rllx)))
    ix2 = minloc(gbl_lon,dim=1,mask=abs(gbl_lon-(rurx-rdx)).eq.minval(abs(gbl_lon-(rurx-rdx))))
    jy1 = minloc(gbl_lat,dim=1,mask=abs(gbl_lat-rlly).eq.minval(abs(gbl_lat-rlly)))
    jy2 = minloc(gbl_lat,dim=1,mask=abs(gbl_lat-(rury-rdy)).eq.minval(abs(gbl_lat-(rury-rdy))))

    print*, 'save_flux: gbl_lon = ',gbl_lon
    print*, 'save_flux: gbl_lat = ',gbl_lat
    print*, 'save_flux: ix1,ix2,jy1,jy2 = ',ix1,ix2,jy1,jy2

    ! space and time conversions
    ! --------------------------

    if ( config%spec.eq.'co2' ) then
      allocate( tmp(nxregrid,nyregrid,nd_nee) )
      ! CO2 species
      jd = juldatei
      d = 0
      do while (jd.le.juldatef)
        d = d + 1
        n = int(floor((jd - juldatei)/real(statres,kind=8))) + 1
        n = min(n,ntstate)
        tmp = flx(ix1:ix2,jy1:jy2,(d-1)*nd_nee+1:d*nd_nee)
        do jy = 1, nyregrid
          do ix = 1, nxregrid
            if ( nbox_xy(ix,jy).le.0 ) cycle
            do nd = 1, nd_nee
              hutc = real(nd - 1)*24./real(nd_nee)
              hh = hutc + hloc(nbox_xy(ix,jy))
              if ( hh.lt.0 ) then
                hh = hh + 24
               else if ( hh.ge.24 ) then
                hh = hh - 24
              endif
              do nn = 1, ndt
                if ( (hh.ge.(nn-1)*24/ndt).and.(hh.lt.nn*24/ndt) ) nt = nn
              end do
              if ( hh.gt.((nt-1)*24/ndt+24/ndt/2).and.nt.lt.ndt ) then
                ! interpolate between current and next timesteps
                flx_out(ix,jy,(d-1)*nd_nee+nd) = state((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) + &
                                      (state((n-1)*ndt*nbox+nt*nbox+nbox_xy(ix,jy)) - &
                                      state((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))) * &
                                      (hh - (nt-1)*24/ndt - 24/ndt/2)*real(ndt)/24. + &
                                      tmp(ix,jy,nd)
              else if ( hh.lt.((nt-1)*24/ndt+24/ndt/2).and.nt.gt.1) then
                ! interpolate between previous and current timesteps 
                flx_out(ix,jy,(d-1)*nd_nee+nd) = state((n-1)*ndt*nbox+(nt-2)*nbox+nbox_xy(ix,jy)) + &
                                      (state((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) - &
                                      state((n-1)*ndt*nbox+(nt-2)*nbox+nbox_xy(ix,jy))) * &
                                      (hh - (nt-2)*24/ndt - 24/ndt/2)*real(ndt)/24. + &
                                      tmp(ix,jy,nd)
              else
                ! take current timestep
                flx_out(ix,jy,(d-1)*nd_nee+nd) = state((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) + &
                                                 tmp(ix,jy,nd)
              endif
            end do
          end do
        end do
        jd = jd + 1d0
      end do
      deallocate( tmp )
    endif ! CO2

    if ( config%spec.eq.'ghg' ) then
      ! GHG species
!      jd = juldatei
!      n = 1
!      do while ( (jd.lt.juldatef).and.(n.lt.nflx) )
      do n = 1, nflx
        nt = minloc(stat_time,dim=1,mask=abs(stat_time-flx_time(n)).eq.minval(abs(stat_time-flx_time(n))))
        nt = min(nt,ntstate)
        print*, 'stat_time(nt), flx_time(n) = ',stat_time(nt), flx_time(n)
        do jy = 1, nyregrid
          do ix = 1, nxregrid
            ! if inc_ocean is true state vector includes all fluxes, if false need to add ocean fluxes
            if ( nbox_xy(ix,jy).gt.0 ) then
              if ( stat_time(nt).lt.flx_time(n).and.(nt.gt.1) ) then
                ! interpolate between previous and current timesteps
                if ( config%lognormal ) then
                  ! lognormal
                  flx_out(ix,jy,n) = flx(ix1+ix-1,jy1+jy-1,n) * ( state((nt-2)*nbox+nbox_xy(ix,jy)) + &
                                     (state((nt-1)*nbox+nbox_xy(ix,jy))-state((nt-2)*nbox+nbox_xy(ix,jy))) * &
                                     (flx_time(n) - stat_time(nt))/statres )
                else
                  ! normal
                  flx_out(ix,jy,n) = state((nt-2)*nbox+nbox_xy(ix,jy)) + &
                                     (state((nt-1)*nbox+nbox_xy(ix,jy))-state((nt-2)*nbox+nbox_xy(ix,jy))) * &
                                     (flx_time(n) - stat_time(nt))/statres + &
                                     flx(ix1+ix-1,jy1+jy-1,n)
                endif
              else if ( stat_time(nt).gt.flx_time(n).and.(nt.lt.ntstate) ) then
                ! interpolate between current and next timesteps
                if ( config%lognormal ) then
                  ! lognormal
                  flx_out(ix,jy,n) = flx(ix1+ix-1,jy1+jy-1,n) * ( state((nt-1)*nbox+nbox_xy(ix,jy)) + &
                                     (state(nt*nbox+nbox_xy(ix,jy))-state((nt-1)*nbox+nbox_xy(ix,jy))) * &
                                     (stat_time(nt) - flx_time(n))/statres )
                else
                  ! normal
                  flx_out(ix,jy,n) = state((nt-1)*nbox+nbox_xy(ix,jy)) + &
                                     (state(nt*nbox+nbox_xy(ix,jy))-state((nt-1)*nbox+nbox_xy(ix,jy))) * &
                                     (stat_time(nt) - flx_time(n))/statres + &
                                     flx(ix1+ix-1,jy1+jy-1,n)
                endif
              else
                ! take current timestep
                if ( config%lognormal ) then
                  flx_out(ix,jy,n) = flx(ix1+ix-1,jy1+jy-1,n) * state((nt-1)*nbox+nbox_xy(ix,jy))
                else
                  flx_out(ix,jy,n) = state((nt-1)*nbox+nbox_xy(ix,jy)) + flx(ix1+ix-1,jy1+jy-1,n)
                endif
              endif
            else
              ! add ocean fluxes if not optimized
              flx_out(ix,jy,n) = flx(ix1+ix-1,jy1+jy-1,n)
            endif
          end do
        end do
      end do
    endif ! GHG

    ! replace fluxes in inversion domain with perturbed fluxes    
    flx(ix1:ix2,jy1:jy2,:) = flx_out(:,:,:)

    ! undo numerical scaling
    flx = flx/numscale
    ! convert to kg/m2/h
    flx = flx*3600.

    print*, 'mod_save: min/max flx = ',minval(flx), maxval(flx)

    ! adjust time to days since 1-Jan-1900
    flx_time = flx_time - juldate(19000101, 0) !+ 1d0

    ! write flux to file
    ! ------------------

    if ( config%spec.eq.'co2' ) then
      if ( config%nested ) then
        varname = files%varnest_nee
        lonname = files%lonnest_nee
        latname = files%latnest_nee
        timename = files%timenest_nee
      else
        varname = files%varname_nee
        lonname = files%lonname_nee
        latname = files%latname_nee
        timename = files%timename_nee
      endif
    else
      if ( config%nested ) then
        varname = files%varnest_flx
        lonname = files%lonnest_flx
        latname = files%latnest_flx
        timename = files%timenest_flx
      else
        varname = files%varname_flx
        lonname = files%lonname_flx
        latname = files%latname_flx
        timename = files%timename_flx
      endif
    endif

    call caldate(juldatei, yyyymmdd, hhmiss)
    write(adate,'(I4.4)') yyyymmdd/10000    
    if ( config%spec.eq.'co2' ) then
      if ( .not.config%nested ) then
        filename = str_replace(files%filename_nee, 'YYYY', adate)
      else
        filename = str_replace(files%filenest_nee, 'YYYY', adate)
      endif
    else
      if ( .not.config%nested) then
        filename = str_replace(files%filename_flx, 'YYYY', adate)
      else
        filename = str_replace(files%filenest_flx, 'YYYY', adate)
      endif
    endif
    call check( nf90_create(trim(files%path_output)//filename, nf90_clobber, ncid) )
    call check( nf90_def_dim(ncid, lonname, nxgrid, lon_dimid) )
    call check( nf90_def_var(ncid, lonname, nf90_real, lon_dimid, lon_varid) )
    call check( nf90_put_att(ncid, lon_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, latname, nygrid, lat_dimid) )
    call check( nf90_def_var(ncid, latname, nf90_real, lat_dimid, lat_varid) )
    call check( nf90_put_att(ncid, lat_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, timename, nflx, time_dimid) )
    call check( nf90_def_var(ncid, timename, nf90_real, time_dimid, time_varid) )
    call check( nf90_put_att(ncid, time_varid, units, timeunit) )
    dimids = (/ lon_dimid, lat_dimid, time_dimid /)
    call check( nf90_def_var(ncid, varname, nf90_real, dimids, varid) )
    call check( nf90_put_att(ncid, varid, units, flxunit) )
    call check( nf90_enddef(ncid) )
    call check( nf90_put_var(ncid, lon_varid, gbl_lon) )
    call check( nf90_put_var(ncid, lat_varid, gbl_lat) )
    call check( nf90_put_var(ncid, time_varid, flx_time) )
    call check( nf90_put_var(ncid, varid, flx) )
    call check( nf90_close(ncid) )

  end subroutine save_flux

  ! --------------------------------------------------
  ! save_obs
  ! --------------------------------------------------
  !> save_obs
  !!
  !! Purpose:  Save the perturbed mixing ratios to
  !!           file in the same format as the 
  !!           observations (these are pseudo-
  !!           observations).
  !!
  ! --------------------------------------------------

  subroutine save_obs(config, files, obs, obserr)

    use mod_var
    use mod_settings

    implicit none

    type (config_t),                  intent (in) :: config
    type (files_t),                   intent (in) :: files
    real, dimension(maxobs),          intent (in) :: obs, obserr

    character(len=max_path_len)                   :: filename
    character(len=max_name_len)                   :: rowfmt, dump
    character(len=recname_len), dimension(nobs)   :: recs
    integer, dimension(nobs)                      :: jjjjmmdd, hhmiss
    real(kind=8), dimension(nobs)                 :: jdate
    real                                          :: conc, cini, cinipos, bkg, nee, fff, ocn, ghg, prior, post, delta, err
    integer                                       :: i, ierr
    logical                                       :: lexist

    ! read monitor
    ! ------------

    filename = trim(files%path_output)//'monitor.txt'
    inquire(file=trim(filename), exist=lexist)
    if ( .not.lexist ) then
      write(logid,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif

    open(100,file=trim(filename),status='old',action='read',iostat=ierr)
    read(100,fmt='(A)') dump
    if ( config%spec.eq.'co2' ) then
      write(rowfmt,'(A,I6,A)') '(A3,1X,I8,1X,I6,1X,F14.6,1X,',11,'(F10.4,1X))'
      do i = 1, nobs
        read(100,fmt=rowfmt) recs(i), jjjjmmdd(i), hhmiss(i), jdate(i), conc, cini, cinipos, bkg, &
                  nee, fff, ocn, prior, post, delta, err
      end do
    else
      write(rowfmt,'(A,I6,A)') '(A3,1X,I8,1X,I6,1X,F14.6,1X,',9,'(F10.4,1X))'
      do i = 1, nobs
        read(100,fmt=rowfmt) recs(i), jjjjmmdd(i), hhmiss(i), jdate(i), conc, cini, cinipos, bkg, &
                  ghg, prior, post, delta, err
      end do
    endif
    close(100)

    ! write obs for each receptor
    ! ---------------------------

    do i = 1, nobs
      filename=trim(files%path_output)//trim(recs(i))//'_'//trim(config%spec)//'.txt'
      inquire(file=trim(filename),exist=lexist)
      if ( .not.lexist ) then
        open(100,file=trim(filename),action='write',status='replace',iostat=ierr)   
      else
        open(100,file=trim(filename),action='write',status='old',access='append',iostat=ierr)   
      endif
      write(100,fmt='(A3,1X,I8,1X,I6,1X,F14.6,1X,F10.4,1X,F10.4,1X,I4)',iostat=ierr) &
              recs(i), jjjjmmdd(i), hhmiss(i), jdate(i), obs(i), obserr(i), 1
      close(100)
    end do

  end subroutine save_obs

  ! --------------------------------------------------
  ! check
  ! --------------------------------------------------

  subroutine check(status)

    integer, intent(in) :: status

    if ( status.ne.nf90_noerr ) then
      write(logid,*) trim(nf90_strerror(status))
      stop
    end if

  end subroutine check

  ! --------------------------------------------------

end module mod_save

