!---------------------------------------------------------------------------------------
! FLEXINVERT: simulate
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> simulate
!!
!! Purpose:    Models the concentrations/mixing ratios using flexpart flux 
!!             sensitivities. Calculates the gradient and cost in the observation space.
!!
!! Interface:
!!
!!    Inputs
!!             iter    -  iteration
!!             files   -  file data structure
!!             config  -  configuration settings data structure
!!             fluxes  -  flux data structure
!!             obs     -  observation data structure
!!             states  -  state vector data structure
!!             grad_o  -  gradient for the observation space (zero on input)
!!             cost_o  -  cost for the observation space (zero on input)
!!
!!    Externals
!!             caldate
!!             get_nread
!!             read_dates
!!             read_grid
!!             read_grid_ncdf
!!             read_factor
!!             unique
!!             calc_conc
!!             save_obs
!!
!---------------------------------------------------------------------------------------

subroutine simulate(iter, files, config, fluxes, obs, states, grad_o, cost_o)

  use mod_flexpart
  use mod_settings
  use mod_var
  use mod_dates
  use mod_fluxes
  use mod_obs
  use mod_states
  use mod_tools
  use mod_save
  use mod_perturb

  implicit none

  integer,                 intent (in)     :: iter
  type (files_t),          intent (in)     :: files
  type (config_t),         intent (in)     :: config
  type (fluxes_t),         intent (in out) :: fluxes
  type (obs_t),            intent (in out) :: obs
  type (states_t),         intent (in)     :: states
  real, dimension(nvar),   intent (in out) :: grad_o
  real,                    intent (in out) :: cost_o

  character(len=max_path_len)              :: path_flexrec, subdir_fp, filename, filenest
  character(len=max_path_len)              :: filedates, filefactor
  character(len=max_path_len)              :: rowfmt
  character(len=6)                         :: adate
  character(len=2)                         :: aiter
  character(len=8)                         :: areldate, areltime
  logical                                  :: lexist
  integer                                  :: ierr
  integer                                  :: jjjjmmdd, hhmiss, hh, hutc
  integer                                  :: month, prevmonth
  integer                                  :: i, it, ix, jy, n, nb, ni, ns, nt, nn
  integer                                  :: ixn1, ixn2, jyn1, jyn2
  integer                                  :: ix1, ix2, jy1, jy2
  integer                                  :: ind, ihi, ilo, flag, ntstep 
  integer                                  :: countrate
  integer                                  :: tloop1, tloop2, time1, time2
  integer                                  :: nread, ngrid 

  integer, dimension(:), allocatable       :: istate, istateuni           ! indices to state vector
  real, dimension(:), allocatable          :: rstateuni                   ! indices to state vector
  real, dimension(nxgrid,nygrid,maxngrid)  :: grid                        ! flexpart flux sensitivity (or footprints)
  real, dimension(nnxgrid,nnygrid,maxngrid):: gridnest                    ! flexpart nest flux sensitivity (or footprints)
  real, dimension(:,:,:), allocatable      :: factor                      ! correction factor for dry air 
  real(kind=8), dimension(maxngrid)        :: gtime                       ! flux sensitivity time stamp (julian days)
  real(kind=8), dimension(:), allocatable  :: ftime                       ! correction factor time stamp (julian days)
  real(kind=8), dimension(:), allocatable  :: dates                       ! correction factor time stamp (jjjjmmddhhmiss)
  real, dimension(:), allocatable          :: hbox                        ! flux sensitivity for boxes
  real, dimension(:,:,:), allocatable      :: hnest                       ! flux sensitivity for regional domain
  real, dimension(nxgrid,nygrid)           :: hbkg                        ! flux sensitivity for global minus regional domain
  real, dimension(:), allocatable          :: px                          ! state variables corresponding to back trajectory time
  real, dimension(:), allocatable          :: hx                          ! transport operator at state vector resolution
!  real, dimension(nxregrid*nyregrid)       :: gridnestvec                 ! vectorized form of grid or gridnest for one timestep

  ! initialize
  ! ----------

  cost_o = 0.
  prevmonth = 0

  ! indices of regional domain
  if ( config%nested ) then
    ixn1 = int((rllx - nllx + ndx)/ndx)
    ixn2 = int((rurx - nllx)/ndx)
    jyn1 = int((rlly - nlly + ndy)/ndy)
    jyn2 = int((rury - nlly)/ndy)
  endif
  ix1 = int((rllx - llx + dx)/dx)
  ix2 = int((rurx - llx)/dx)
  jy1 = int((rlly - lly + dy)/dy)
  jy2 = int((rury - lly)/dy)

  write(logid,*) 'simulate: indices regional domain = ',ix1,ix2,jy1,jy2
  if ( config%nested ) write(logid,*) 'simulate: indices regional domain nest = ',ixn1,ixn2,jyn1,jyn2

  ! loop over observations
  ! ----------------------

  do i = 1, nobs

    call system_clock(tloop1, countrate)
  
    ! transport operators
    ! -------------------

    ! define file names
    call caldate(obs%jdate(i), jjjjmmdd, hhmiss)
    month = jjjjmmdd/100
    write(adate,'(I6.6)') jjjjmmdd/100
    write(areldate,'(I8.8)') jjjjmmdd
    write(areltime,'(I6.6)') hhmiss

    ! read correction factor for dry dir (ratio rho_wet/rho_dry)
!    if ( month.ne.prevmonth ) then
!      if ( allocated(dates) ) deallocate(dates)
!      if ( allocated(ftime) ) deallocate(ftime)
!      if ( allocated(factor) ) deallocate(factor)
!      call get_nread(filedates, nread)
!      allocate ( dates(nread) )
!      allocate ( ftime(nread) )
!      allocate ( factor(nxgrid,nygrid,nread) )
!      call read_dates(filedates, nread, dates)
!      do n = 1, nread
!        ftime(n) = juldate(int(dates(n)/1d6),int(dates(n)-floor(dates(n)/1d6)*1d6)) 
!      end do
!      factor(:,:,:) = 1. 
!      filefactor = trim(path_flexrec)//'factor_drygrid'
!      inquire ( file=trim(filefactor),exist=lexist )
!      if ( lexist ) then
!        write(logid,*) 'Reading correction factor :'//trim(filefactor)
!        call read_factor(filefactor, nread, nxgrid, nygrid, nxshift, factor)
!      endif
!    endif 
    ! for nested fp run: assume dry air correction is negligible

    ! read flexpart footprints
    call system_clock(time1, countrate)
    if ( .not.config%average_fp ) then
      ! footprint times match observation timestamps
      path_flexrec = trim(files%path_flexpart)//trim(obs%recs(i))//'/'//trim(adate)//'/'
      filename = trim(path_flexrec)//'grid_time_'//trim(areldate)//trim(areltime)//'_001'
      filenest = trim(path_flexrec)//'grid_time_nest_'//trim(areldate)//trim(areltime)//'_001'
      filedates = trim(path_flexrec)//'dates'
      inquire ( file=trim(filename),exist=lexist )
      if ( lexist ) then
        write(logid,*) 'Reading flexpart :'//trim(filename)
        ! read footprints from binary files
        call read_grid(filename, filedates, obs%jdate(i), nxgrid, nygrid, nxshift, ngrid, grid, gtime)
      else
        write(logid,*) 'WARNING: cannot find '//trim(filename)
        go to 10
      endif
    else
      ! footprints averaged/selected to match observation timestamps 
      subdir_fp = trim(obs%recs(i))//'/'//trim(adate)//'/'
      filename = trim(files%path_flexncdf)//trim(subdir_fp)//'grid_time_'//trim(areldate)//trim(areltime)//'_001.nc'
      inquire ( file=trim(filename),exist=lexist )
      if ( lexist ) then
        write(logid,*) 'Reading flexpart :'//trim(filename)
        ! read averaged footprints from ncdf files
        call read_grid_ncdf(filename, nxgrid, nygrid, ngrid, grid, gtime)
      else
        write(logid,*) 'WARNING: cannot find '//trim(filename)
        go to 10
      endif
    endif
    call system_clock(time2, countrate)
    if (i.lt.10.and.iter.eq.0) write(logid,*) 'simulate: time to read grid (ms) = ',(time2-time1)*1000/countrate

    ! correction for dry air (only if obs are mixing ratios) 
!    do n = 1, ngrid
!      ind = minloc(ftime,dim=1,mask=abs(ftime-gtime(n)).eq.minval(abs(ftime-gtime(n))))
!      grid(:,:,n) = grid(:,:,n)*factor(:,:,ind)
!    end do
    ! convert from ppt to ppmv or ppbv
    grid = grid*config%coeff*mmair/config%molarmass
    ! convert s.m3/kg to s.m2/kg
    grid = grid/outheight(1)
    ! apply numerical scaling
    grid = grid/numscale

    ! read nested flexpart footprints
    if ( config%nested ) then
      if ( .not.config%average_fp ) then
        ! footprint times match observation timestamps
        filenest = trim(path_flexrec)//'grid_time_nest_'//trim(areldate)//trim(areltime)//'_001'
        inquire ( file=trim(filenest),exist=lexist )
        if ( lexist ) then
          write(logid,*) 'Reading flexpart :'//trim(filenest)
          ! read footprints from binary files
          call read_grid(filenest, filedates, obs%jdate(i), nnxgrid, nnygrid, nnxshift, ngrid, gridnest, gtime)
        else
          write(logid,*) 'WARNING: cannot find '//trim(filenest)
          go to 10
        endif
      else
        ! footprints averaged/selected to match observation timestamps 
        subdir_fp = trim(obs%recs(i))//'/'//trim(adate)//'/'
        filenest = trim(files%path_flexncdf)//trim(subdir_fp)//'grid_time_nest_'//trim(areldate)//trim(areltime)//'_001.nc'
        inquire ( file=trim(filenest),exist=lexist )
        if ( lexist ) then
          ! read averaged footprints from ncdf files
          call read_grid_ncdf(filenest, nnxgrid, nnygrid, ngrid, gridnest, gtime)
        else
          write(logid,*) 'WARNING: cannot find '//trim(filenest)
          go to 10
        endif
      endif
      ! convert from ppt to ppmv or ppbv
      gridnest = gridnest*config%coeff*mmair/config%molarmass
      ! convert s.m3/kg to s.m2/kg
      gridnest = gridnest/outheight(1)
      ! apply numerical scaling
      gridnest = gridnest/numscale
    endif

    ! calculate indices to state vector for footprint times
    if ( .not.allocated(istate) ) allocate( istate(ngrid) )
    if ( .not.allocated(istateuni) ) allocate( istateuni(ngrid) )
    if ( .not.allocated(rstateuni) ) allocate( rstateuni(ngrid) )
    do n = 1, ngrid
      ni = 0
      if ( gtime(n).lt.(states%xtime(ntstate)+real(statres,kind=8)) ) then 
        flag = 0
      else
        flag = 1
      endif
      do while ( flag.eq.0 )
        ni = ni + 1
        if ( gtime(n).lt.(states%xtime(ni)+real(statres,kind=8)) ) flag = 1
      end do
      istate(n) = ni
    end do
    call unique(ngrid, real(istate), rstateuni, ntstep)
    istateuni = int(rstateuni)
    if ( config%verbose ) then
      write(logid,*) 'simulate: istate = ',istate
      write(logid,*) 'simulate: ntstep = ',ntstep
      write(logid,*) 'simulate: istateuni = ',istateuni(1:ntstep)
    endif
    if ( ntstep.eq.0 ) go to 10

    ! transport for boxes in regional domain
    call system_clock(time1, countrate)
    if ( .not.allocated(hbox) ) allocate( hbox(nbox*ngrid) )
    hbox(:) = 0.
    do n = 1, ngrid
      do jy = 1, nyregrid
        do ix = 1, nxregrid
          if ( nbox_xy(ix,jy).gt.0 ) then
            if ( config%nested ) then
              hbox((n-1)*nbox+nbox_xy(ix,jy)) = hbox((n-1)*nbox+nbox_xy(ix,jy)) + gridnest(ixn1+ix-1,jyn1+jy-1,n)
            else
              hbox((n-1)*nbox+nbox_xy(ix,jy)) = hbox((n-1)*nbox+nbox_xy(ix,jy)) + grid(ix1+ix-1,jy1+jy-1,n)
            endif
          endif
        end do
      end do
    end do
    call system_clock(time2, countrate)
    if (i.lt.10.and.iter.eq.0) write(logid,*) 'simulate: time to calc hbox (ms) = ',(time2-time1)*1000/countrate

    ! transport for grid regional domain 
    ! to calculate contribution of fixed fluxes in domain
    if ( .not.allocated(hnest) ) allocate( hnest(nxregrid,nyregrid,ngrid) )
    if ( config%nested ) then
      hnest(:,:,:) = gridnest(ixn1:ixn2,jyn1:jyn2,1:ngrid)
    else
      hnest(:,:,:) = grid(ix1:ix2,jy1:jy2,1:ngrid)
    endif

    ! transport operator for background contribution
    ! no influence from inside the domain
    hbkg(:,:) = sum(grid(:,:,1:ngrid),dim=3) 
    hbkg(ix1:ix2,jy1:jy2) = 0.

    ! transport matrix for analytical method
    if ( config%method.eq.'analytic' ) then
      do n = 1, ngrid
        ind = istate(n)
        if ( (ind.eq.0).or.(ind.gt.ntstate) ) cycle
        if ( config%spec.ne.'co2' ) then
          hmat(i,(ind-1)*nbox+1:ind*nbox) = hmat(i,(ind-1)*nbox+1:ind*nbox) + hbox((n-1)*nbox+1:n*nbox)
        else
          hutc = int((gtime(n) - floor(gtime(n)))*24)
          do nb = 1, nbox
            hh = hutc + int(hloc(nb))
            if ( hh.lt.0 ) then
              hh = hh + 24
            else if ( hh.ge.24 ) then
              hh = hh - 24
            endif
            do nn = 1, ndt
              if ( (hh.ge.(nn-1)*24/ndt).and.(hh.lt.nn*24/ndt) ) nt = nn
            end do
            hmat(i,(ind-1)*ndt*nbox+(nt-1)*nbox+nb) = hmat(i,(ind-1)*ndt*nbox+(nt-1)*nbox+nb) + hbox((n-1)*nbox+nb)
          end do
        endif
      end do
    endif

    ! species specific preparation
    ! ----------------------------

    call system_clock(time1, countrate)
    if (config%spec.eq.'co2') then

      ! CO2 tracer 
      ! ----------

      if ( config%method.eq.'congrad'.or.config%method.eq.'m1qn3' ) then
        ! select state variables for current footprint times
        allocate( px(ntstep*ndt*nbox), stat=ierr )
        if( ierr.ne.0 ) then
          write(logid,*) 'ERROR simulate: not enough memory'
          stop
        endif
        allocate( hx(ntstep*ndt*nbox), stat=ierr )
        if( ierr.ne.0 ) then
          write(logid,*) 'ERROR simulate: not enough memory'
          stop
        endif
        hx(:) = 0.
        do n = 1, ngrid
          if( istate(n).eq.0 ) cycle
          ns = istate(n) - minval(istate,mask=(istate.gt.0)) + 1
          ! state vector, px 
          px((ns-1)*ndt*nbox+1:ns*ndt*nbox) = states%px((istate(n)-1)*ndt*nbox+1:istate(n)*ndt*nbox)
          ! transport operator, hx
          hutc = int((gtime(n) - floor(gtime(n)))*24)
          do nb = 1, nbox
            hh = hutc + int(hloc(nb))
            if ( hh.lt.0 ) then
              hh = hh + 24
            else if ( hh.ge.24 ) then
              hh = hh - 24
            endif
            do nn = 1, ndt
              if ( (hh.ge.(nn-1)*24/ndt).and.(hh.lt.nn*24/ndt) ) nt = nn
            end do
            hx((ns-1)*ndt*nbox+(nt-1)*nbox+nb) = hx((ns-1)*ndt*nbox+(nt-1)*nbox+nb) + hbox((n-1)*nbox+nb)
          end do
        end do
      endif
 
    else

      ! non-CO2 tracer
      ! --------------

      if ( config%method.eq.'congrad'.or.config%method.eq.'m1qn3' ) then
        ! select state variables for current footprint times and 
        ! average transport operator to state variable temporal resolution
        allocate( px(ntstep*nbox), stat=ierr )
        if( ierr.ne.0 ) then
          write(logid,*) 'ERROR simulate: not enough memory'
          stop
        endif
        allocate( hx(ntstep*nbox), stat=ierr )
        if( ierr.ne.0 ) then
          write(logid,*) 'ERROR simulate: not enough memory'
          stop
        endif
        hx(:) = 0.
        do n = 1, ngrid
          if ( istate(n).eq.0 ) cycle
          ns = istate(n) - minval(istate,mask=(istate.gt.0)) + 1
          ! state vector, px
          if ( config%lognormal ) then
            ! lognormal
            ! optimize scalar z = ln(x/xb) so x = exp(z)*xb
            px((ns-1)*nbox+1:ns*nbox) = exp(states%px((istate(n)-1)*nbox+1:istate(n)*nbox)) * &
                                               fluxes%flx_box(:,istate(n),1)
          else
            ! normal
            px((ns-1)*nbox+1:ns*nbox) = states%px((istate(n)-1)*nbox+1:istate(n)*nbox)
          endif
          ! transport operator, hx
          hx((ns-1)*nbox+1:ns*nbox) = hx((ns-1)*nbox+1:ns*nbox) + hbox((n-1)*nbox+1:n*nbox)
        end do
      endif

    endif
    call system_clock(time2, countrate)
    if (i.lt.10.and.iter.eq.0) write(logid,*) 'simulate: time to calc hx and px (ms) = ',(time2-time1)*1000/countrate

    ! verbose output
    ! --------------

    if ( config%verbose ) then
      ! hbox
      filename = trim(files%path_output)//'hbox_'//trim(areldate)//trim(areltime)//'.txt'
      open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
      do n = 1, nbox*ngrid
        write(100,fmt='(E11.4)') hbox(n)
      end do
      close(100)
      ! hnest
      filename = trim(files%path_output)//'hnest_'//trim(areldate)//trim(areltime)//'.txt'
      open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
      write(rowfmt,'(A,I6,A)') '(',nxregrid,'(E11.4,1X))'
      do n = 1, ngrid
        do jy = 1, nyregrid
          write(100,fmt=rowfmt) hnest(:,jy,n)
        end do
      end do
      close(100)
    endif 

    ! model mixing ratios
    ! -------------------

    if ( iter.eq.0.or.(config%restart.eq.1.and.iter.eq.(lastiter+1)) ) then
      call system_clock(time1, countrate)
      ! compute fixed mixing ratios
      call calc_conc(config, fluxes, obs, ngrid, gtime, hnest, hbkg, i, ix1, ix2, jy1, jy2)
      call system_clock(time2, countrate)
      if (i.eq.1.and.iter.eq.0) write(logid,*) 'simulate: time to calc conc (ms) = ',(time2-time1)*1000/countrate
    endif ! iter

    ! contribution from state vector fluxes
    if ( config%method.eq.'congrad'.or.config%method.eq.'m1qn3' ) then
      obs%model(i) = dot_product(hx, px)
    else
      obs%model(i) = dot_product(hmat(i,:), states%px)
    endif

    ! contribution from state vector scalars of initial mixing ratio
    if ( config%opt_cini ) then
      ! ni time step of initial mixing ratio optimization
      ni = 0
      do while (.true.)
        ni = ni + 1
        if ( ni.gt.ntcini ) exit
        if ( (obs%jdate(i)-trajdays).lt.(states%cinitime(ni)+real(cinires,kind=4)) ) exit
      end do
      ni = min(ntcini, ni)
      if ( config%method.eq.'congrad'.or.config%method.eq.'m1qn3' ) then
        if ( config%lognormal ) then
          ! lognormal 
          ! x = xb*exp(z) where xb = 1
          obs%cinipos(i) = dot_product(obs%cini(i,:), exp(states%px(npvar+(ni-1)*ncini+1:npvar+ni*ncini)))
        else
          ! normal
          obs%cinipos(i) = dot_product(obs%cini(i,:), states%px(npvar+(ni-1)*ncini+1:npvar+ni*ncini))
        endif
      else
        hmat(i,npvar+(ni-1)*ncini+1:npvar+ni*ncini) = obs%cini(i,:)
        obs%cinipos(i) = dot_product(hmat(i,npvar+(ni-1)*ncini+1:npvar+ni*ncini), states%px(npvar+(ni-1)*ncini+1:npvar+ni*ncini))
      endif
    else
      obs%cinipos(i) = sum(obs%cini(i,:))
    endif

    ! store prior value
    if ( iter.eq.0 ) obs%prior(i) = obs%model(i)

    ! model-observation difference
    ! ydel = Hf(p) + Hf_fix - yobs + yini + ybkg
    if (config%spec.eq.'co2') then
      obs%delta(i) = obs%model(i) + obs%nee(i) + obs%fff(i) + obs%ocn(i) - &
                      obs%conc(i) + obs%bkg(i) + obs%cinipos(i)
    else
      if ( config%lognormal ) then
        ! optimize scalar of fluxes so no contribution from best guess fluxes (ghg) 
        ! but add ocn contribution (zero if ocean pixels optimized)
        obs%delta(i) = obs%model(i) - obs%conc(i) + obs%bkg(i) + obs%cinipos(i) + obs%ocn(i)
      else
        ! optimize offsets so account for contribution from best guess fluxes (ghg)
        obs%delta(i) = obs%model(i) + obs%ghg(i) - obs%conc(i) + obs%bkg(i) + obs%cinipos(i)
      endif
    endif

    ! if background missing then observation should have no influence! 
!    if ( sum(obs%cini(i,:)).eq.0. ) then
!      write(logid,*) 'WARNING: background zero for observation ',i
!      obs%delta(i) = 0.
!    endif

    ! calculate gradient in observation space
    ! ---------------------------------------

    call system_clock(time1, countrate)
    if ( config%method.eq.'congrad'.or.config%method.eq.'m1qn3' ) then
      do n = 1, ntstep
        if ( istateuni(n).gt.0 ) then 
          if ( config%lognormal ) then
            ! lognormal
            ! Jo'(p) = diag(p0*exp(p))H^TR^-1(H(xb*exp(p)) - y)
            ! calculate as: Jo'(p) = sum( xb*exp(p)*H_i^T*ydel_i*R_i ) (note here px = xb*exp(p))
            grad_o((istateuni(n)-1)*ndvar+1:istateuni(n)*ndvar) = grad_o((istateuni(n)-1)*ndvar+1:istateuni(n)*ndvar) + &
                                                                    px((n-1)*ndvar+1:n*ndvar) * hx((n-1)*ndvar+1:n*ndvar) * obs%delta(i) / obs%err(i)**2
          else
            ! normal
            ! Jo'(p) = H^TR^-1(H(p) - y)
            ! calculate as: Jo'(p) = sum( H_i^T*ydel_i*R_i )
            grad_o((istateuni(n)-1)*ndvar+1:istateuni(n)*ndvar) = grad_o((istateuni(n)-1)*ndvar+1:istateuni(n)*ndvar) + &
                                                                    hx((n-1)*ndvar+1:n*ndvar) * obs%delta(i) / obs%err(i)**2
          endif
        endif
      end do
    endif

    ! calculate contribution to gradient from initial mixing ratios
    if ( config%opt_cini ) then
      if ( config%lognormal ) then
        ! lognormal
        grad_o(npvar+(ni-1)*ncini+1:npvar+ni*ncini) = grad_o(npvar+(ni-1)*ncini+1:npvar+ni*ncini) + &
                                                        exp(states%px(npvar+(ni-1)*ncini+1:npvar+ni*ncini)) * obs%cini(i,:) * obs%delta(i) / obs%err(i)**2
      else
        ! normal
        grad_o(npvar+(ni-1)*ncini+1:npvar+ni*ncini) = grad_o(npvar+(ni-1)*ncini+1:npvar+ni*ncini) + &
                                                        obs%cini(i,:) * obs%delta(i) / obs%err(i)**2
      endif
    endif
    call system_clock(time2, countrate)
    if (i.lt.10.and.iter.eq.0) write(logid,*) 'simulate: time to calc grad (ms) = ',(time2-time1)*1000/countrate

    ! cost observation space
    ! ----------------------

    ! 2Jo = (y - Hp)^TR^(-1)(y - Hp)
    cost_o = cost_o + obs%delta(i)**2/obs%err(i)**2

10 continue

    if ( allocated(px) )        deallocate(px)
    if ( allocated(hx) )        deallocate(hx)
    if ( allocated(hbox) )      deallocate(hbox)
    if ( allocated(hnest) )     deallocate(hnest)
    if ( allocated(istate) )    deallocate(istate)
    if ( allocated(istateuni) ) deallocate(istateuni)
    if ( allocated(rstateuni) ) deallocate(rstateuni)

    prevmonth = month
    call system_clock(tloop2, countrate)
    if (i.lt.10.and.iter.eq.0) write(logid,*) 'simulate: time one loop (ms) = ',(tloop2-tloop1)*1000/countrate

  end do  ! nobs

  ! save output
  ! -----------

  call save_obs(files, config, obs)

  if (config%method.eq.'analytic') then
    ! save hmat
    write(rowfmt,'(A,I6,A)') '(',nvar,'(E11.4,1X))' 
    filename = trim(files%path_output)//'hmat.txt'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)  
    do i = 1, nobs
      write(100,fmt=rowfmt) hmat(i,:)
    end do
    close(100)
  else
    ! save gradient
    write(aiter,'(I2.2)') iter
    filename = trim(files%path_output)//'grad_'//aiter//'.txt'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
    do i = 1, nvar
!      write(100,fmt='(E11.4)') grad_o(i)
      write(100,fmt='(E13.6)') grad_o(i)
    end do
    close(100)
  endif

  ! prepare to exit
  ! ---------------

  if ( allocated(dates) )  deallocate(dates)
  if ( allocated(ftime) )  deallocate(ftime)
  if ( allocated(factor) ) deallocate(factor)
 
end subroutine simulate


